//arduino pro or pro mini
#include <stdlib.h>
#include <Math.h>
#include <LiquidCrystal.h>

const int RS = 22, EN = 24, D4 = 25, D5 = 26, D6 = 27, D7 = 28;
LiquidCrystal lcd(RS, EN, D4, D5, D6, D7);

char gCmdBuffer[512];
unsigned char gCmdBufferPoint=0;
char* gLook=NULL;

//make locals globals for nonrecursive fxns to get good idea of how much dynamic mem is used.

     
unsigned int gCmdID=0;
double gNumber=0.0;
char gName[32];
byte gErrorState=0;

const unsigned char XYSTEPPERENABLE=42;
const unsigned char ZSTEPPERENABLE=43;
const unsigned char XAXISDIR=47;
const unsigned char XAXIS=46;
const unsigned char ZAXISDIR=45;
const unsigned char ZAXIS=44;
const unsigned char YAXISDIR=49;
const unsigned char YAXIS=48;

const unsigned char XY_MS1=52;
const unsigned char XY_MS2=53;
const unsigned char Z_MS1=50;
const unsigned char Z_MS2=51;

const unsigned char PADLEFT=A2;
const unsigned char PADRIGHT=A5;
const unsigned char PADZUP=A1;
const unsigned char PADZDOWN=A0;
const unsigned char PADUP=A4;
const unsigned char PADDOWN=A3;
const unsigned char PADENTER=A6;

const unsigned char FANPIN=13;
const unsigned char LEVELPIN=A7;

const byte AXISSTEP[]={XAXIS,YAXIS,ZAXIS};
const byte AXISDIR[]={XAXISDIR,YAXISDIR,ZAXISDIR};

const unsigned int STEPS_PER_XY[2]={40,1016};//0 - mm, 1 - in     Eighth stepping
const unsigned int STEPS_PER_Z[2]={100,2540};//0 - mm, 1 - in     Quarter stepping
const double MAX_FEED_RATE[2]={8000,314};//0-mm, 1 - in

long gCurPos[3]={0,0,0};//current x, y, and z positions in step increments.
long gZActualPos=0;//current Z position during auto level operations

byte gKeyPadEnabled=0;
byte gAutoLevelOn=0;

const long BLOCKS_MAX_W=30;
const long BLOCKS_MAX_H=30;
long BLOCKSW=0;
long BLOCKSH=0;
long BLOCKW=1;
long BLOCKH=1;
short gALHighestOffset=0;
short ZDATA[BLOCKS_MAX_W+1][BLOCKS_MAX_H+1];

double gXWord,gYWord,gZWord,gDwellWord,gEWord,gSWord,gTWord,gBWord,gWWord,gHWord,gQWord,gRWord;

const byte METRIC=0;
const byte IMPERIAL=1;
byte gUnitType=METRIC;

double gFeedWord=MAX_FEED_RATE[gUnitType];

const byte ABSOLUTE=0;
const byte RELATIVE=1;
byte gCoordType=ABSOLUTE;

const byte UNITSPERMIN=0;
const byte INVERSETIME=1;
byte gFeedType=UNITSPERMIN;

void updateLCD(void){
    static double lastX=-12.34;
    static double lastY=-12.34;
    static double lastZ=-12.34;
    static byte lastKeyPadEnabled=12;
    static byte lastAutoLevelOn=12;
    static byte lastErrorState=12;
    static byte lastUnitType=12;
    double x=(double)gCurPos[0]/STEPS_PER_XY[gUnitType];
    double y=(double)gCurPos[1]/STEPS_PER_XY[gUnitType];
    double z=(double)gCurPos[2]/STEPS_PER_Z[gUnitType];
    
    if (x!=lastX){
        lastX=x;
        lcd.setCursor(0,0);
        lcd.print("X:      ");
        lcd.setCursor(3,0);
        lcd.print(x);
    }
    if (y!=lastY){
        lastY=y;
        lcd.setCursor(8,0);
        lcd.print("Y:      ");
        lcd.setCursor(11,0);
        lcd.print(y);
    }
    if (z!=lastZ){
        lastZ=z;
        lcd.setCursor(0,1);
        lcd.print("Z:      ");
        lcd.setCursor(3,1);
        lcd.print(z);
    }

    if (!gErrorState){
        if (gUnitType!=lastUnitType || gErrorState!=lastErrorState){
            lastUnitType=gUnitType;
            lcd.setCursor(8,1);
            if (gUnitType==METRIC){
                lcd.print("mm");
            }else{
                lcd.print("in");
            }
        }
        if (gAutoLevelOn!=lastAutoLevelOn || gErrorState!=lastErrorState){
            lastAutoLevelOn=gAutoLevelOn;
            lcd.setCursor(10,1);
            if (gAutoLevelOn){
                lcd.print("AL");
            }else{
                lcd.print("  ");
            }
        }
        if (gKeyPadEnabled!=lastKeyPadEnabled || gErrorState!=lastErrorState){
            lastKeyPadEnabled=gKeyPadEnabled;
            lcd.setCursor(12,1);
            if (gKeyPadEnabled){
                lcd.print("key");
            }else{
                lcd.print("   ");
            }
        }
    }    
    if (gErrorState!=lastErrorState){
        lastErrorState=gErrorState;
        if (gErrorState){
            lcd.setCursor(8,1);
            lcd.print("ERRORFLG");
        }
    }
}

void error(const __FlashStringHelper* msg){
     Serial.print(F("error: "));
     Serial.println(msg);
     gErrorState=1;
     updateLCD();
}

unsigned char inline isWhiteSpace(char c){
     return (c<=32 && c>=1);
}
unsigned char inline isAlpha(char c){
  return ((c>='a' && c<='z') || (c>='A' && c<='Z'));
}
unsigned char inline isDigit(char c){
  return (c>='0' && c<='9'); 
}
unsigned char inline isAlNum(char c){
  return isDigit(c) || isAlpha(c); 
}
char inline toLower(char c){
     if (c>='A' && c<='Z') c-='A'+'a';
     return c;
}

void inline getChar(void){  
     if (*gLook!=0 && *gLook!=';'){
          gLook++;
     }else if (*gLook==';'){
          *gLook=0; 
     }
}

void skipWhite(void){ 
     while (isWhiteSpace(*gLook)){
          getChar(); 
     }
}

void getRealNumber(void){
     char numBuff[20];
     char sign=1;
     byte isfloat=0;
     int numLen=0;

     if (*gLook=='-'){
          sign=-1;
          getChar();
     }else if (*gLook=='+'){
          getChar(); 
     }
     while (numLen<sizeof(numBuff)-1){
          if (isDigit(*gLook) || (isfloat==0 && *gLook=='.')){
               if (*gLook=='.') isfloat=1;
               numBuff[numLen++]=*gLook;  
               getChar();
          }else{
               break; 
          }
     }
     if ((numBuff[0]=='.' && numLen>1) || (numLen>0 && numBuff[0]!='.')){
          numBuff[numLen]=0; 
          gNumber=atof(numBuff)*sign;
          skipWhite();
     }else{
          error(F("Expected number"));
          return;
     }
}
void getCmdID(void){
    char cmdBuff[20];
    short cmdLen=0;
    while (cmdLen<sizeof(cmdBuff)-1){
        if (!isDigit(*gLook)) break;
        cmdBuff[cmdLen++]=*gLook; 
        getChar(); 
    }
    if (cmdLen>0){
        cmdBuff[cmdLen]=0;    
        gCmdID=atoi(cmdBuff);
        skipWhite();
    }else{
        error(F("Expected cmd id"));
        return;
    }
}

void getName(void){
    int nameLen=0;
    if (isAlpha(*gLook)==0){
        error(F("Expected name"));  
        return;
    }
    while (isAlNum(*gLook)){
        if (nameLen<sizeof(gName)-1){
            gName[nameLen++]=*gLook;
            getChar();
        }else{
            gName[0]=0;
            error(F("Name buffer overflow"));
            return;
        }
    }
    gName[nameLen]=0;
    skipWhite(); 
}

unsigned char getParityByte(char* string){
  int len=strlen(string);
  unsigned char parity=0;
  for (int i=0;i<strlen(string);i++){
    parity+=string[i];
  }
  if (parity==0 || parity==10 || parity==13){
    parity+=100;
  }
  return parity;
}

double inline GetDistance(double x1,double y1,double z1,double x2,double y2,double z2){
    return sqrt(fabs((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1)));
}

void delayUSeconds(unsigned int t){
     if (t>16000){
          while (t>16000){
               delayMicroseconds(16000);
               t-=16000; 
          }
          delayMicroseconds(t);
     }else{
          delayMicroseconds(t); 
     }
}

void enableMotors(void){
  digitalWrite(FANPIN,1);
  digitalWrite(XYSTEPPERENABLE,0);//enable motors
  digitalWrite(ZSTEPPERENABLE,0);  
}

void disableMotors(void){
  digitalWrite(XYSTEPPERENABLE,1);//disable motors
  digitalWrite(ZSTEPPERENABLE,1);
  digitalWrite(FANPIN,0); 
}


void cncInit(void){
  pinMode(FANPIN,OUTPUT);
  digitalWrite(FANPIN,0);
  
  pinMode(XYSTEPPERENABLE,OUTPUT);
  pinMode(ZSTEPPERENABLE,OUTPUT);
  digitalWrite(XYSTEPPERENABLE,1);
  digitalWrite(ZSTEPPERENABLE,1);
  pinMode(XAXIS, OUTPUT);
  pinMode(XAXISDIR, OUTPUT);
  pinMode(YAXIS, OUTPUT);
  pinMode(YAXISDIR, OUTPUT);
  pinMode(ZAXIS, OUTPUT);
  pinMode(ZAXISDIR, OUTPUT);
  
  pinMode(PADLEFT,INPUT);
  pinMode(PADRIGHT,INPUT);
  pinMode(PADZUP,INPUT);
  pinMode(PADZDOWN,INPUT);
  pinMode(PADUP,INPUT);
  pinMode(PADDOWN,INPUT);
  pinMode(PADENTER,INPUT);
  
  digitalWrite(PADLEFT,1);
  digitalWrite(PADRIGHT,1);
  digitalWrite(PADZUP,1);
  digitalWrite(PADZDOWN,1);
  digitalWrite(PADUP,1);
  digitalWrite(PADDOWN,1);
  digitalWrite(PADENTER,1);

  pinMode(XY_MS1,OUTPUT);
  pinMode(XY_MS2,OUTPUT);
  pinMode(Z_MS1,OUTPUT);
  pinMode(Z_MS2,OUTPUT);

  digitalWrite(XY_MS1,1);
  digitalWrite(XY_MS2,1);
  digitalWrite(Z_MS1,0);
  digitalWrite(Z_MS2,1);
  
  pinMode(LEVELPIN,INPUT);
  digitalWrite(LEVELPIN,1);

}

void inline doSteps(void){ 
     delayMicroseconds(50);
     digitalWrite(XAXIS,0);
     digitalWrite(YAXIS,0);
     digitalWrite(ZAXIS,0);
}

void inline setDir(byte which, char dir){
     if (dir>0){
          digitalWrite(AXISDIR[which],1);
     }else if (dir<0){
          digitalWrite(AXISDIR[which],0);
     }
}

double BilinearInterpolation(double x,double y, double x2,double y2, double TL, double TR, double BL, double BR, double PX, double PY){
  double R1=abs((PX-x)*(PY-y));
  double R2=abs((x2-PX)*(PY-y));
  double R3=abs((PX-x)*(y2-PY));
  double R4=abs((x2-PX)*(y2-PY));
  double TotalArea=R1+R2+R3+R4;
  double R1p=R4/TotalArea;
  double R2p=R3/TotalArea;
  double R3p=R2/TotalArea;
  double R4p=R1/TotalArea;
  return (R1p*TL)+(R2p*TR)+(R3p*BL)+(R4p*BR);
}

long ZOffset(long x,long y){
    long wx=x/BLOCKW;
    long wy=y/BLOCKH;
    if (wx>BLOCKSW-2 || wy>BLOCKSH-2 || x<0 || y<0) return gALHighestOffset;
    long TL=ZDATA[wx][wy];
    long TR=ZDATA[wx+1][wy];
    long BL=ZDATA[wx][wy+1];
    long BR=ZDATA[wx+1][wy+1];
    return BilinearInterpolation(wx*BLOCKW,wy*BLOCKH,(wx+1)*BLOCKW,(wy+1)*BLOCKH,TL,TR,BL,BR,x,y);
}

void gotoOneAxisSteps(byte which,long to,unsigned int sleepTime){
    if (gCurPos[which]>to){
        setDir(which,-1);
        for (;gCurPos[which]!=to;gCurPos[which]--){
            digitalWrite(AXISSTEP[which],1);
            delayMicroseconds(50);
            digitalWrite(AXISSTEP[which],0);
            delayUSeconds(sleepTime);
        }
    }else if (gCurPos[which]<to){
        setDir(which,1);
        for (;gCurPos[which]!=to;gCurPos[which]++){
            digitalWrite(AXISSTEP[which],1);
            delayMicroseconds(50);
            digitalWrite(AXISSTEP[which],0);
            delayUSeconds(sleepTime);
        }
    }
}

void gotoXYZSteps(long toX,long toY,long toZ,byte absRel,byte unitType,byte feedType,double feedRate,byte autoLevelOn, byte smoothingOn){
    long delta[3];
    long finalPos[3];
    long error[2]={0,0};
    char axisList[3]={0,1,2};
    char inc[3]={1,1,1};
    double distance;
    long tempSwapLong;
    char tempSwapChar;
    long countdown;
    long lineDev;
    long zDelta;
    char zInc;
    char zNormalInc;
    char i,j;
    long zDiff;
    unsigned int uiSleepTime;
    long accelCounter=0;
    
    if (absRel==ABSOLUTE){
        delta[0]=toX-gCurPos[0];
        delta[1]=toY-gCurPos[1];
        delta[2]=toZ-gCurPos[2];
    }else if (absRel==RELATIVE){
        delta[0]=toX;
        delta[1]=toY;
        delta[2]=toZ;
    }

    if (delta[0]==0 && delta[1]==0 && delta[2]==0) return;//no movement, dont waste clocks on nothing

    distance=GetDistance(0,0,0,(double)delta[0]/STEPS_PER_XY[unitType],(double)delta[1]/STEPS_PER_XY[unitType],(double)delta[2]/STEPS_PER_Z[unitType]); 

    for (i=0;i<3;i++){//determines move directions and double deltas
        finalPos[i]=gCurPos[i]+delta[i];//record final positions to make the steps match in the end
        if (delta[i]<0){
            inc[i]=-1;
            delta[i]=abs(delta[i]);
        }
        setDir(axisList[i],inc[i]);
        delta[i]<<=1;
    }
    zNormalInc=inc[2];
    
    for (j=0;j<3;j++){//bubble sort
        for (i=0;i<2;i++){
            if (delta[i]<delta[i+1]){
                tempSwapLong=delta[i];
                delta[i]=delta[i+1];
                delta[i+1]=tempSwapLong;
                tempSwapChar=axisList[i];
                axisList[i]=axisList[i+1];
                axisList[i+1]=tempSwapChar;
                tempSwapChar=inc[i];
                inc[i]=inc[i+1];
                inc[i+1]=tempSwapChar;
            }      
        }
    }

    if (feedType==UNITSPERMIN){
        uiSleepTime=min(max(300,(60.0f/(feedRate/distance)*1000000.0f)/((double)delta[0]/2)),65000);
    }else{
        uiSleepTime=min(max(300,(feedRate/((double)delta[0]/2))*1000000.0f),65000);
    }
    if (autoLevelOn){
        uiSleepTime-=250;
    }
    
    lineDev=delta[0];
    countdown=delta[0]>>1;
    while (countdown--){
        zDiff=gCurPos[2];
        gCurPos[axisList[0]] += inc[0];//step the dominate axis
        digitalWrite(AXISSTEP[axisList[0]],1);
        
        if (error[0] > lineDev){//step another axis
            gCurPos[axisList[1]] += inc[1];
            error[0] -= delta[0];
            digitalWrite(AXISSTEP[axisList[1]],1);
        }
        error[0]+=delta[1];
        
        if (error[1] > lineDev){//and step the last one
            gCurPos[axisList[2]] += inc[2];
            error[1] -= delta[0];
            digitalWrite(AXISSTEP[axisList[2]],1);
        }
        error[1]+=delta[2];
        
        doSteps();
        if (autoLevelOn){
            zDiff=gCurPos[2]-zDiff;
            gZActualPos+=zDiff;
            zDelta=gCurPos[2]+ZOffset(gCurPos[0],gCurPos[1])-gZActualPos;
            if (zDelta){
                if (zDelta<0){
                    zInc=-1;
                }else{
                    zInc=1;
                }
                setDir(2,zInc);
                while (zDelta!=0){
                    zDelta=zDelta-zInc;
                    gZActualPos+=zInc;
                    digitalWrite(ZAXIS,1);
                    delayMicroseconds(50);
                    digitalWrite(ZAXIS,0);
                    delayUSeconds(uiSleepTime);
                }
                setDir(2,zNormalInc);
            }else{
              delayUSeconds(uiSleepTime);
            }
        }else{
            delayUSeconds(uiSleepTime);
        }
        
        if (accelCounter<25 && smoothingOn){
            delayMicroseconds((125-accelCounter)*5);
            accelCounter++;
        }else if (countdown<25 && smoothingOn){
            delayMicroseconds((125-countdown)*5);
        }
    }

    if (autoLevelOn) uiSleepTime+=250;//clean up straggler steps to make it so the position was the original destination
    for (i=0;i<3;i++){
        gotoOneAxisSteps(i,finalPos[i],uiSleepTime);
    }
    
    updateLCD();
}

void gotoXYZ(double toX, double toY, double toZ,byte absRel,byte unitType,byte feedType,double feedRate,byte autoLevelOn,byte smoothingOn){
  gotoXYZSteps(toX*(long)STEPS_PER_XY[unitType] , toY*(long)STEPS_PER_XY[unitType] , toZ*(long)STEPS_PER_Z[unitType] , absRel , unitType , feedType , feedRate , autoLevelOn , smoothingOn);
}

void zeroZAxis(double maxDepth){
    long maxDepthSteps=maxDepth*STEPS_PER_Z[gUnitType];
    byte foundOffset=0;
    for (long zoff=gCurPos[2];zoff>maxDepthSteps;zoff--){
        if (!digitalRead(LEVELPIN)){
          delay(25);
          if (!digitalRead(LEVELPIN)){
              foundOffset=1;
              break;
          }
        }
        gotoOneAxisSteps(2,zoff,1050);
    }
    if (!foundOffset){
        error(F("Couldnt find Z zero. There is a crash potential."));
        return;
    }
    gCurPos[2]=0;
    if (gAutoLevelOn){
        gZActualPos=ZOffset(gCurPos[0],gCurPos[1]);
    }
    updateLCD();
}

void buildAutoLevelData(double w, double h, long columns, long rows, double probeStartHeight,double probeDeepestHeight){
    long probeX=0;
    long probeY=0;
    long probeZStart=probeStartHeight*STEPS_PER_Z[gUnitType];
    long probeZEnd=probeDeepestHeight*STEPS_PER_Z[gUnitType];
    long zoff=0;
    byte foundOffset=0;
    if (columns<2) columns=2;
    if (columns>BLOCKS_MAX_W) columns=BLOCKS_MAX_W;
    if (rows<2) rows=2;
    if (rows>BLOCKS_MAX_H) rows=BLOCKS_MAX_H;
    if (probeZStart<probeZEnd){
        error(F("Probe depth is higher then probe start height"));
        return;
    }
    gALHighestOffset=0;
    BLOCKSW=columns;
    BLOCKSH=rows;
    BLOCKW=(w*STEPS_PER_XY[gUnitType])/(columns-1);
    BLOCKH=(h*STEPS_PER_XY[gUnitType])/(rows-1);

    gotoXYZSteps(gCurPos[0],gCurPos[1],probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
    gotoXYZSteps(0,0,probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
    zeroZAxis(probeDeepestHeight);
    ZDATA[0][0]=0;
  
    gotoXYZSteps(gCurPos[0],gCurPos[1],probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
    for (int y=0;y<BLOCKSH;y++){
        probeX=0;
        for (int x=0;x<BLOCKSW;x++){
            if (x==0 && y==0){
                //Already zeroed the Z axis to 0,0 so no need to redo work
            }else{
                gotoXYZSteps(probeX,probeY,probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
                for (zoff=probeZStart;zoff>probeZEnd;zoff--){
                    if (!digitalRead(LEVELPIN)){
                      delay(25);
                      if (!digitalRead(LEVELPIN)){
                          foundOffset=1;
                          break;
                      }
                    }
                    gotoOneAxisSteps(2,zoff,1050);
                }
                if (!foundOffset){
                    error(F("Couldnt find Z offset. There is a crash potential."));
                    return;
                }
                foundOffset=0;
                ZDATA[x][y]=zoff;
                if (zoff>gALHighestOffset) gALHighestOffset=zoff;
            }
            gotoXYZSteps(probeX,probeY,probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
            probeX+=BLOCKW;
        }
        probeY+=BLOCKH;
    }
    gotoXYZSteps(0,0,probeZStart,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
    gotoXYZSteps(0,0,0,ABSOLUTE,METRIC,UNITSPERMIN,MAX_FEED_RATE[METRIC],0,1);
}

void checkKeypad(void){//Let the operator move errthang around to, makes it easier to zero everything.
    static unsigned int keypadSpeed=0;
    if (gKeyPadEnabled){
         byte moved=0;
         float x=0;
         float y=0;
         float z=0;
         if (!digitalRead(PADUP)){
            y+=0.1;
            moved=1;
          }
          if (!digitalRead(PADDOWN)){
            y-=0.1;
            moved=1;
          }
          if (!digitalRead(PADLEFT)){
            x-=0.1;
            moved=1;
          }
          if (!digitalRead(PADRIGHT)){
            x+=0.1;
            moved=1;
          }
          if (!digitalRead(PADZUP)){
            z+=0.1;
            moved=1;
          }
          if (!digitalRead(PADZDOWN)){
            z-=0.1;
            moved=1;
          }
          if (moved){
              if (keypadSpeed<5000){
                keypadSpeed+=5;
              }
              gotoXYZ(x,y,z,RELATIVE,METRIC,UNITSPERMIN,200+keypadSpeed,gAutoLevelOn,0); 
          }else{
               keypadSpeed=0; 
          }
     }
}

byte inline isWord(char c){
     switch (toLower(c)){
          case 'x':
          case 'y':
          case 'z':
          case 'f':
          case 'p':
          case 's':
          case 't':
          case 'b':
          case 'w':
          case 'h':
          case 'e':
          case 'q':
          case 'r':
               return 1;
     } 
     return 0;
}

byte loadWord(char wordType){
    if (!isWord(wordType)){
         return 0; 
    }
    getRealNumber();
    switch (toLower(wordType)){
         case 'x': 
              gXWord=gNumber;
              break;
         case 'y':
              gYWord=gNumber;
              break;
         case 'z':
              gZWord=gNumber;
              break;
         case 'f':
              gFeedWord=gNumber;
              break;
         case 'p':
              gDwellWord=gNumber;
              break;
         case 's':
              gSWord=gNumber;
              break;
         case 't':
              gTWord=gNumber;
              break;
         case 'b':
              gBWord=gNumber;
              break;
         case 'w':
              gWWord=gNumber;
              break;
         case 'h':
              gHWord=gNumber;
              break;
         case 'e':
              gEWord=gNumber;
              break;
         case 'q':
              gQWord=gNumber;
              break;
         case 'r':
              gRWord=gNumber;
              break;
    }
    return 1;
}
void loadWords(void){
     char wordType;
     while (isWord(*gLook)){
          wordType=*gLook;
          getChar();
          skipWhite();
          if (!loadWord(wordType)){
               break;
          }
     }
}

byte setWord(char wordType){
    if (!isWord(wordType)){
         return 0; 
    }
    getRealNumber();
    switch (toLower(wordType)){
         case 'x': 
              gXWord=gNumber;
              gCurPos[0]=gNumber*(long)STEPS_PER_XY[gUnitType];
              break;
         case 'y':
              gYWord=gNumber;
              gCurPos[1]=gNumber*(long)STEPS_PER_XY[gUnitType];
              break;
         case 'z':
              gZWord=gNumber;
              gCurPos[2]=gNumber*(long)STEPS_PER_Z[gUnitType];
              break;
         default:
              error(F("Unknown G92 word"));
              return 0;
    }
    return 1;
}

void setWords(void){
     char wordType;
     while (isWord(*gLook)){
          wordType=*gLook;
          getChar();
          skipWhite();
          if (!setWord(wordType)){
               break;
          }
     }
    updateLCD();
}

void interpret(void){
  static char cmdType;
  long tempPos;
  gLook=&gCmdBuffer[0];
  if (*gLook==';') *gLook=0;//the whole line is a comment, null it out and foget bout it.
  skipWhite();
     while (*gLook!=0 && *gLook!=';'){
          cmdType=toLower(*gLook);
          getChar();
          skipWhite();
          switch (cmdType){
               case 'g':
                    if (gErrorState==0){
                        getCmdID();
                        if (gCmdID!=92){//If its 92, dont load all the words, we'll set the words in the case 92.
                          loadWords();
                        }
                        switch (gCmdID){
                             case 0:
                                  gotoXYZ(gXWord,gYWord,gZWord,gCoordType,gUnitType,UNITSPERMIN,MAX_FEED_RATE[gUnitType],gAutoLevelOn,1);
                                  break;  
                             case 1:
                                  gotoXYZ(gXWord,gYWord,gZWord,gCoordType,gUnitType,gFeedType,gFeedWord,gAutoLevelOn,1);
                                  break;  
                             case 4:
                                  delay(gDwellWord*1000);
                                  break;  
                             case 20:
                                  if (gUnitType==METRIC && gFeedType==UNITSPERMIN){
                                      gFeedWord=gFeedWord/25.4;
                                  }
                                  gUnitType=IMPERIAL;
                                  updateLCD();
                                  break;  
                             case 21:
                                  if (gUnitType==IMPERIAL && gFeedType==UNITSPERMIN){
                                      gFeedWord=gFeedWord*25.4;
                                  }
                                  gUnitType=METRIC;
                                  updateLCD();
                                  break;  
                             case 90:
                                  gCoordType=ABSOLUTE;
                                  break;  
                             case 91:
                                  gCoordType=RELATIVE;
                                  break;
                             case 92:
                                  setWords();
                                  break;  
                             case 93:
                                  gFeedType=INVERSETIME;
                                  break;  
                             case 94:
                                  gFeedType=UNITSPERMIN;
                                  break;
                             default:
                                  error(F("Unsupported G cmd id"));
                                  break;
                        }
                    }else{
                         error(F("Clear error flag before any more commands, use m105")); 
                    }
                    break;
               case 'm':
                    getCmdID();
                    loadWords();
                    switch (gCmdID){
                         case 101://Enable motors
                              enableMotors(); 
                              break;
                         case 102://Disable motors
                              disableMotors();
                              break;
                         case 103://Enable keypad
                              gKeyPadEnabled=1;
                              updateLCD();
                              break;
                         case 104://Disable keypad
                              gKeyPadEnabled=0;
                              updateLCD();
                              break;
                         case 105://Clear error state
                              gErrorState=0;
                              updateLCD();
                              break;
                         case 300://probe work peice and build autolevel data
                              if (gAutoLevelOn){
                                  error(F("Cannot build auto level table while auto level is active."));
                                  break;
                              }
                              buildAutoLevelData(gWWord, gHWord, gQWord, gRWord, gTWord, gBWord);
                              break;
                         case 301://zero Z axis to work peice
                              zeroZAxis(gBWord);
                              break;
                         case 310://Enable autolevel
                              if (!gAutoLevelOn){
                                  gAutoLevelOn=1;
                                  gZActualPos=gCurPos[2]+ZOffset(gCurPos[0],gCurPos[1]);
                                  tempPos=gCurPos[2];
                                  gotoXYZSteps(gCurPos[0],gCurPos[1],gZActualPos,ABSOLUTE,gUnitType,gFeedType,gFeedWord,0,1);
                                  gCurPos[2]=tempPos;
                              }
                              updateLCD();
                              break;
                         case 311://Disable autolevel
                              if (gAutoLevelOn){
                                  gAutoLevelOn=0;
                                  tempPos=gCurPos[2];
                                  gCurPos[2]=gZActualPos;
                              }
                              updateLCD();
                              break;
				                 case 200://Wait for user to hit enter to resume
                               Serial.println(F("PAUSED, PRESS ENTER TO RESUME"));
                               while (1){
                                if (!digitalRead(PADENTER)){
                                  delay(50);
                                  if (!digitalRead(PADENTER)){
                                    break;
                                  }
                                }                                
                                checkKeypad();
                               }                                
                               while (!digitalRead(PADENTER)){
                               }
                               Serial.println(F("ENTER PRESSED, RESUMING"));
                        	     delay(500);
                               break;
                         default:
                              error(F("Unsupported G cmd id"));
                              break;
	     
                    }
                    //TODO: find out what M codes I care about and include them
                    break;
               default:           
                    if (!loadWord(cmdType)){
                         error(F("Unrecognized command"));
                    } 
                    break;
          }
          if (gErrorState){
               break; 
          }
     }
}

void loop() {
     static unsigned long lastReadySent=0;
     static unsigned char parityByte=0;
     checkKeypad();
     if (Serial.available()>0){
          int readByte=Serial.read();
          if (readByte!=-1){
               if (readByte==10 || readByte==13){
                    if (gCmdBufferPoint>0){
                         gCmdBuffer[gCmdBufferPoint]=0;
                         parityByte=gCmdBuffer[0];
                         if (getParityByte(&gCmdBuffer[1])==parityByte){
                              Serial.print('k');
                              gCmdBuffer[0]=' ';
                              interpret();
                              Serial.print("r\n");
                              lastReadySent=millis();
                              gCmdBufferPoint=0;
                         }else{
                              gCmdBufferPoint=0;
                              Serial.print('b'); 
                         }
                    }else{
                         Serial.print("r\n");
                         lastReadySent=millis();
                    }

               }else{
                    if (gCmdBufferPoint<sizeof(gCmdBuffer)-1){
                         gCmdBuffer[gCmdBufferPoint++]=readByte;
                    }else{
                         error(F("command buffer overflow"));
                         gCmdBufferPoint=0;
                    }
               }
         }
     }
     if ((millis()-lastReadySent>1000 || millis()<lastReadySent)){//ensure a ready is sent at least every 1000ms
          Serial.print("r\n"); 
          lastReadySent=millis(); 
     }
}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(250000);
  cncInit();

  digitalWrite(23,0);
  pinMode(23,OUTPUT);
  lcd.begin(16, 2);
  updateLCD();
  Serial.print("r\n");
}
